@@ -9,6 +9,8 @@ import subprocess
 import urllib
 import zc.buildout
 
+from string import strip
+
 class Recipe(object):
     """zc.buildout recipe for compiling and installing software"""
 
@@ -29,6 +31,7 @@ class Recipe(object):
 
         self.gems = options['gems'].split()
         self.version = options.get('version')
+        self.url = options.get('url')
         # Allow to define specific ruby executable. If not, take just 'ruby'
         self.ruby_executable = options.get('ruby-executable', 'ruby')
 
@@ -59,13 +62,22 @@ class Recipe(object):
     def _join_paths(self, *paths):
         return ':'.join(filter(None, paths))
 
+    def _get_env_override(self, env):
+        env = filter(None, map(strip, env.splitlines()))
+        try:
+            env = [map(strip, line.split('=', 1)) for line in env]
+        except ValueError: # Unpacking impossible
+            self.log.error("Every environment line should contain a '=' sign")
+            zc.buildout.UserError('Configuration error')
+        return env
+
     def _get_env(self):
         s = {
             'PATH': os.environ.get('PATH', ''),
             'PREFIX': self.options['location'],
             'RUBYLIB': os.environ.get('RUBYLIB', ''),
         }
-        return {
+        env = {
             'GEM_HOME': '%(PREFIX)s/lib/ruby/gems/1.8' % s,
             'RUBYLIB': self._join_paths(
                     '%(RUBYLIB)s',
@@ -78,14 +90,25 @@ class Recipe(object):
                     '%(PREFIX)s/bin',
                 ) % s,
         }
+        env_override = self.options.get('environment', '')
+        env_override = self._get_env_override(env_override)
+        env.update({k: v % env for k, v in env_override})
+        return env
 
     def _get_latest_rubygems(self):
+        if self.url:
+            version = self.version
+            if not version:
+                version = re.search(r'rubygems-([0-9.]+).zip$', self.url).group(1)
+            return (self.url, version)
+
         if self.version:
             return ('http://production.cf.rubygems.org/rubygems/'
                     'rubygems-%s.zip' % self.version, self.version)
 
         f = urllib.urlopen('http://rubygems.org/pages/download')
         s = f.read()
+        s = unicode(s)
         f.close()
         r = re.search(r'http://production.cf.rubygems.org/rubygems/'
                       r'rubygems-([0-9.]+).zip', s)
@@ -136,7 +159,7 @@ class Recipe(object):
     def _install_executable(self, path):
         content = ['#!/bin/sh']
         for key, val in self._get_env().items():
-            content.append('export %s=%s' % (key, val))
+            content.append("export %s='%s'" % (key, val))
         content.append('%s "$@"' % path)
         name = os.path.basename(path)
         bindir = self.buildout['buildout']['bin-directory']
@@ -165,6 +188,8 @@ class Recipe(object):
             gem_executable = self.get_gem_executable(bindir)
 
         for gemname in self.gems:
+            extra = self.options.get('gem-options', '')
+            extra = ' '.join(filter(None, map(strip, extra.splitlines())))
             s = {
                 'GEM': gem_executable,
                 'OPTIONS': ' '.join([
@@ -172,14 +197,15 @@ class Recipe(object):
                         '--no-ri',
                         '--bindir=%s' % bindir,
                     ]),
+                'EXTRA': extra,
             }
             if '==' in gemname:
-                gemname, version = gemname.split('==', 1)
-                s['GEMNAME'] = gemname.strip()
-                s['OPTIONS'] += ' --version %s' % version.strip()
+                gemname, version = map(strip, gemname.split('==', 1))
+                s['GEMNAME'] = gemname
+                s['OPTIONS'] += ' --version %s' % version
             else:
                 s['GEMNAME'] = gemname
-            self.run('%(GEM)s install %(OPTIONS)s %(GEMNAME)s' % s,
+            self.run('%(GEM)s install %(OPTIONS)s %(GEMNAME)s -- %(EXTRA)s' % s,
                      self._get_env())
 
         for executable in os.listdir(bindir):
